@@ -40,13 +40,15 @@
 #include <unistd.h>
 #include <sys/types.h>
 #include <sys/stat.h>
-#include <time.h>
+#include <errno.h>
 
 #include <nbdkit-plugin.h>
 
+#ifndef O_CLOEXEC
+#define O_CLOEXEC 0
+#endif
+
 static char *filename = NULL;
-static int rdelayms = 0;        /* read delay (milliseconds) */
-static int wdelayms = 0;        /* write delay (milliseconds) */
 
 static void
 file_unload (void)
@@ -54,32 +56,6 @@ file_unload (void)
   free (filename);
 }
 
-static int
-parse_delay (const char *value)
-{
-  size_t len = strlen (value);
-  int r;
-
-  if (len > 2 && strcmp (&value[len-2], "ms") == 0) {
-    if (sscanf (value, "%d", &r) == 1)
-      return r;
-    else {
-      nbdkit_error ("cannot parse rdelay/wdelay milliseconds parameter: %s",
-                    value);
-      return -1;
-    }
-  }
-  else {
-    if (sscanf (value, "%d", &r) == 1)
-      return r * 1000;
-    else {
-      nbdkit_error ("cannot parse rdelay/wdelay seconds parameter: %s",
-                    value);
-      return -1;
-    }
-  }
-}
-
 /* Called for each key=value passed on the command line.  This plugin
  * only accepts file=<filename>, which is required.
  */
@@ -88,19 +64,15 @@ file_config (const char *key, const char *value)
 {
   if (strcmp (key, "file") == 0) {
     /* See FILENAMES AND PATHS in nbdkit-plugin(3). */
+    free (filename);
     filename = nbdkit_absolute_path (value);
     if (!filename)
       return -1;
   }
-  else if (strcmp (key, "rdelay") == 0) {
-    rdelayms = parse_delay (value);
-    if (rdelayms == -1)
-      return -1;
-  }
-  else if (strcmp (key, "wdelay") == 0) {
-    wdelayms = parse_delay (value);
-    if (wdelayms == -1)
-      return -1;
+  else if (strcmp (key, "rdelay") == 0 ||
+           strcmp (key, "wdelay") == 0) {
+    nbdkit_error ("add --filter=delay on the command line");
+    return -1;
   }
   else {
     nbdkit_error ("unknown parameter '%s'", key);
@@ -118,14 +90,16 @@ file_config_complete (void)
     nbdkit_error ("you must supply the file=<FILENAME> parameter after the plugin name on the command line");
     return -1;
   }
+  if (access (filename, F_OK) < 0) {
+    nbdkit_error ("access '%s': %m", filename);
+    return -1;
+  }
 
   return 0;
 }
 
 #define file_config_help \
-  "file=<FILENAME>     (required) The filename to serve.\n" \
-  "rdelay=<NN>[ms]                Read delay in seconds/milliseconds.\n" \
-  "wdelay=<NN>[ms]                Write delay in seconds/milliseconds." \
+  "file=<FILENAME>     (required) The filename to serve." \
 
 /* The per-connection handle. */
 struct handle {
@@ -171,7 +145,7 @@ file_close (void *handle)
   free (h);
 }
 
-#define THREAD_MODEL NBDKIT_THREAD_MODEL_SERIALIZE_REQUESTS
+#define THREAD_MODEL NBDKIT_THREAD_MODEL_PARALLEL
 
 /* Get the file size. */
 static int64_t
@@ -185,6 +159,19 @@ file_get_size (void *handle)
     return -1;
   }
 
+  if (S_ISBLK (statbuf.st_mode)) {
+    off_t size;
+
+    /* Block device, so st_size will not be the true size. */
+    size = lseek (h->fd, 0, SEEK_END);
+    if (size == -1) {
+      nbdkit_error ("lseek (to find device size): %m");
+      return -1;
+    }
+    return size;
+  }
+
+  /* Else regular file. */
   return statbuf.st_size;
 }
 
@@ -194,14 +181,6 @@ file_pread (void *handle, void *buf, uint32_t count, uint64_t offset)
 {
   struct handle *h = handle;
 
-  if (rdelayms > 0) {
-    const struct timespec ts = {
-      .tv_sec = rdelayms / 1000,
-      .tv_nsec = (rdelayms * 1000000) % 1000000000
-    };
-    nanosleep (&ts, NULL);
-  }
-
   while (count > 0) {
     ssize_t r = pread (h->fd, buf, count, offset);
     if (r == -1) {
@@ -226,14 +205,6 @@ file_pwrite (void *handle, const void *buf, uint32_t count, uint64_t offset)
 {
   struct handle *h = handle;
 
-  if (wdelayms > 0) {
-    const struct timespec ts = {
-      .tv_sec = wdelayms / 1000,
-      .tv_nsec = (wdelayms * 1000000) % 1000000000
-    };
-    nanosleep (&ts, NULL);
-  }
-
   while (count > 0) {
     ssize_t r = pwrite (h->fd, buf, count, offset);
     if (r == -1) {
@@ -248,6 +219,41 @@ file_pwrite (void *handle, const void *buf, uint32_t count, uint64_t offset)
   return 0;
 }
 
+/* Write data to the file. */
+static int
+file_zero (void *handle, uint32_t count, uint64_t offset, int may_trim)
+{
+#if defined(FALLOC_FL_PUNCH_HOLE) || defined(FALLOC_FL_ZERO_RANGE)
+  struct handle *h = handle;
+#endif
+  int r = -1;
+
+#ifdef FALLOC_FL_PUNCH_HOLE
+  if (may_trim) {
+    r = fallocate (h->fd, FALLOC_FL_PUNCH_HOLE | FALLOC_FL_KEEP_SIZE,
+		   offset, count);
+    if (r == -1 && errno != EOPNOTSUPP) {
+      nbdkit_error ("zero: %m");
+    }
+    /* PUNCH_HOLE is older; if it is not supported, it is likely that
+       ZERO_RANGE will not work either, so fall back to write. */
+    return r;
+  }
+#endif
+
+#ifdef FALLOC_FL_ZERO_RANGE
+  r = fallocate (h->fd, FALLOC_FL_ZERO_RANGE, offset, count);
+  if (r == -1 && errno != EOPNOTSUPP) {
+    nbdkit_error ("zero: %m");
+  }
+#else
+  /* Trigger a fall back to writing */
+  errno = EOPNOTSUPP;
+#endif
+
+  return r;
+}
+
 /* Flush the file to disk. */
 static int
 file_flush (void *handle)
@@ -275,7 +281,9 @@ static struct nbdkit_plugin plugin = {
   .get_size          = file_get_size,
   .pread             = file_pread,
   .pwrite            = file_pwrite,
+  .zero              = file_zero,
   .flush             = file_flush,
+  .errno_is_preserved = 1,
 };
 
 NBDKIT_REGISTER_PLUGIN(plugin)
