@@ -51,9 +51,37 @@
 
 #include <nbdkit-plugin.h>
 
+/* XXX Apparently global state is technically wrong in Python 3, see:
+ *
+ * https://www.python.org/dev/peps/pep-3121/
+ *
+ * However it probably doesn't matter for nbdkit since we don't ever
+ * have multiple Python interpreters or multiple instances of the
+ * plugin in a single process.
+ */
 static const char *script;
 static PyObject *module;
 
+static int last_error;
+
+static PyObject *
+set_error (PyObject *self, PyObject *args)
+{
+  int err;
+
+  if (!PyArg_ParseTuple(args, "i", &err))
+    return NULL;
+  nbdkit_set_error (err);
+  last_error = err;
+  Py_RETURN_NONE;
+}
+
+static PyMethodDef NbdkitMethods[] = {
+  { "set_error", set_error, METH_VARARGS,
+    "Store an errno value prior to throwing an exception" },
+  { NULL }
+};
+
 /* Is a callback defined? */
 static int
 callback_defined (const char *name, PyObject **obj_rtn)
@@ -87,10 +115,29 @@ check_python_failure (const char *callback)
   return 0;
 }
 
+#if PY_MAJOR_VERSION >= 3
+static struct PyModuleDef moduledef = {
+  PyModuleDef_HEAD_INIT,
+  "nbdkit",
+  NULL,
+  0,
+  NbdkitMethods,
+  NULL,
+  NULL,
+  NULL,
+  NULL
+};
+#endif
+
 static void
 py_load (void)
 {
   Py_Initialize ();
+#if PY_MAJOR_VERSION >= 3
+  PyModule_Create (&moduledef);
+#else
+  Py_InitModule("nbdkit", NbdkitMethods);
+#endif
 }
 
 static void
@@ -102,6 +149,21 @@ py_unload (void)
   Py_Finalize ();
 }
 
+static void
+py_dump_plugin (void)
+{
+  PyObject *fn;
+  PyObject *r;
+
+  if (callback_defined ("dump_plugin", &fn)) {
+    PyErr_Clear ();
+
+    r = PyObject_CallObject (fn, NULL);
+    Py_DECREF (fn);
+    Py_DECREF (r);
+  }
+}
+
 static int
 py_config (const char *key, const char *value)
 {
@@ -133,7 +195,11 @@ py_config (const char *key, const char *value)
     /* Note that because closeit flag == 1, fp is now closed. */
 
     /* The script should define a module called __main__. */
+#ifdef HAVE_PYSTRING_FROMSTRING
     modname = PyString_FromString ("__main__");
+#else
+    modname = PyUnicode_FromString ("__main__");
+#endif
     module = PyImport_Import (modname);
     Py_DECREF (modname);
     if (!module) {
@@ -154,8 +220,13 @@ py_config (const char *key, const char *value)
     PyErr_Clear ();
 
     args = PyTuple_New (2);
+#ifdef HAVE_PYSTRING_FROMSTRING
     PyTuple_SetItem (args, 0, PyString_FromString (key));
     PyTuple_SetItem (args, 1, PyString_FromString (value));
+#else
+    PyTuple_SetItem (args, 0, PyUnicode_FromString (key));
+    PyTuple_SetItem (args, 1, PyUnicode_FromString (value));
+#endif
     r = PyObject_CallObject (fn, args);
     Py_DECREF (fn);
     Py_DECREF (args);
@@ -413,6 +484,48 @@ py_trim (void *handle, uint32_t count, uint64_t offset)
   return 0;
 }
 
+static int
+py_zero (void *handle, uint32_t count, uint64_t offset, int may_trim)
+{
+  PyObject *obj = handle;
+  PyObject *fn;
+  PyObject *args;
+  PyObject *r;
+
+  if (callback_defined ("zero", &fn)) {
+    PyErr_Clear ();
+
+    last_error = 0;
+    args = PyTuple_New (4);
+    Py_INCREF (obj); /* decremented by Py_DECREF (args) */
+    PyTuple_SetItem (args, 0, obj);
+    PyTuple_SetItem (args, 1, PyLong_FromUnsignedLongLong (count));
+    PyTuple_SetItem (args, 2, PyLong_FromUnsignedLongLong (offset));
+    PyTuple_SetItem (args, 3, PyBool_FromLong (may_trim));
+    r = PyObject_CallObject (fn, args);
+    Py_DECREF (fn);
+    Py_DECREF (args);
+    if (last_error == EOPNOTSUPP) {
+      /* When user requests this particular error, we want to
+         gracefully fall back, and to accomodate both a normal return
+         and an exception. */
+      nbdkit_debug ("zero requested falling back to pwrite");
+      if (r)
+        Py_DECREF (r);
+      PyErr_Clear ();
+      return -1;
+    }
+    if (check_python_failure ("zero") == -1)
+      return -1;
+    Py_DECREF (r);
+    return 0;
+  }
+
+  nbdkit_debug ("zero missing, falling back to pwrite");
+  nbdkit_set_error (EOPNOTSUPP);
+  return -1;
+}
+
 static int
 py_can_write (void *handle)
 {
@@ -437,7 +550,7 @@ py_can_write (void *handle)
     Py_DECREF (r);
     return ret;
   }
-  /* No Perl can_write callback, but there's a Perl pwrite callback
+  /* No Python can_write callback, but there's a Python pwrite callback
    * defined, so return 1.  (In C modules, nbdkit would do this).
    */
   else if (callback_defined ("pwrite", NULL))
@@ -470,7 +583,7 @@ py_can_flush (void *handle)
     Py_DECREF (r);
     return ret;
   }
-  /* No Perl can_flush callback, but there's a Perl flush callback
+  /* No Python can_flush callback, but there's a Python flush callback
    * defined, so return 1.  (In C modules, nbdkit would do this).
    */
   else if (callback_defined ("flush", NULL))
@@ -531,7 +644,7 @@ py_can_trim (void *handle)
     Py_DECREF (r);
     return ret;
   }
-  /* No Perl can_trim callback, but there's a Perl trim callback
+  /* No Python can_trim callback, but there's a Python trim callback
    * defined, so return 1.  (In C modules, nbdkit would do this).
    */
   else if (callback_defined ("trim", NULL))
@@ -552,6 +665,7 @@ static struct nbdkit_plugin plugin = {
 
   .load              = py_load,
   .unload            = py_unload,
+  .dump_plugin       = py_dump_plugin,
 
   .config            = py_config,
   .config_complete   = py_config_complete,
@@ -570,6 +684,7 @@ static struct nbdkit_plugin plugin = {
   .pwrite            = py_pwrite,
   .flush             = py_flush,
   .trim              = py_trim,
+  .zero              = py_zero,
 };
 
 NBDKIT_REGISTER_PLUGIN(plugin)
