@@ -33,15 +33,14 @@
 
 set -e
 set -x
+source ./functions.sh
 
 # Test nbdkit starts up, forks in the background, writes a PID file,
 # and can be killed.
 
 rm -f start.pid start.sock
 
-../src/nbdkit \
-    -P start.pid -U start.sock \
-    ../plugins/example1/.libs/nbdkit-example1-plugin.so
+nbdkit -P start.pid -U start.sock example1
 
 # We may have to wait a short time for the pid file to appear.
 for i in `seq 1 10`; do
